--
-- PCIE TGT Wrapper
--
-- Wraps AVNET PCIe core and buffers data with FIFOs
--
-- Core is a very basic FIFO interface


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

-- synthesis translate_off
library unisim;
use unisim.vcomponents.all;
-- synthesis translate_on

library work;
use work.fifo.all;
use work.fifo_if_types.all;
use work.slave_bus_if_types.all;

entity pcietgt_wrapper_b128 is
  generic (
    family : family_t := family_virtex4);
  port (
    fifo_clk : in std_logic;
    -- 2 x FIFO Interfaces connected to 2 PCIe channels
    rfifo0_inputs    : in  fifo_read_port_inputs_type;
    rfifo0_outputs   : out fifo_read_port_outputs_type;
    wfifo1_inputs    : in  fifo_write_port_inputs_type;
    wfifo1_outputs   : out fifo_write_port_outputs_type;
    -- PCIe ID
    identifier       : in  std_logic_vector(31 downto 0);
    slave_bus_in            : in  slave_bus_outputs_type;
    slave_bus_out           : out slave_bus_inputs_type;
    -- PCIe Interface
    pcie_tx_p        : out std_logic_vector(7 downto 0);
    pcie_tx_n        : out std_logic_vector(7 downto 0);
    pcie_rx_p        : in  std_logic_vector(7 downto 0);
    pcie_rx_n        : in  std_logic_vector(7 downto 0);
    pcie_clk_p         : in  std_logic;
    pcie_clk_n         : in  std_logic;
    pcie_rst_n       : in  std_logic);
end pcietgt_wrapper_b128;

architecture rtl of pcietgt_wrapper_b128 is

  component slimbridge
    port (
      pci_exp_txp  : out std_logic_vector(7 downto 0);
      pci_exp_txn  : out std_logic_vector(7 downto 0);
      pci_exp_rxp  : in  std_logic_vector(7 downto 0);
      pci_exp_rxn  : in  std_logic_vector(7 downto 0);
      sys_clk_p   : in std_logic;
      sys_clk_n   : in std_logic;
      sys_reset_n : in std_logic;
      reg_addr   : in std_logic_vector(7 downto 0);
      reg_data   : out std_logic_vector(31 downto 0);
      id         : in std_logic_vector(31 downto 0);
      fifo_clk   : out std_logic;
      fifo_rst   : out std_logic;
      tofpga_fifo_data  : out std_logic_vector(63 downto 0);
      tofpga_fifo_empty : out std_logic;
      tofpga_fifo_read  : in  std_logic;
      tohost_fifo_data  : in  std_logic_vector(63 downto 0);
      tohost_fifo_empty : in  std_logic;
      tohost_fifo_read  : out std_logic);
  end component;

  signal tofpga_fifo_data,tohost_fifo_data : std_logic_vector(63 downto 0);
  signal tofpga_fifo_empty,tofpga_fifo_read : std_logic;
  signal tohost_fifo_empty,tohost_fifo_read : std_logic;
  
  
  signal tofpga_fifo_data1,tohost_fifo_data1 : std_logic_vector(63 downto 0);
  signal tofpga_fifo_empty1,tofpga_fifo_read1 : std_logic;
  signal tohost_fifo_empty1,tohost_fifo_read1 : std_logic;

  signal ssf0_full,ssf1_full : std_logic;
  
  signal fifo0_nfull_reg : std_logic;
  signal fifo0_ld_iq_reg : std_logic_vector(63 downto 0);
  signal fifo0_ld_iq : std_logic_vector(127 downto 0);
  signal fifo0_write : std_logic;
  signal fifo0_write_q : std_logic;
  signal fifo0_write_toggle : std_logic := '0';
  signal fifo0_nf : std_logic;
  
  signal afifo1_read : std_logic;
  signal afifo1_qdata : std_logic_vector(127 downto 0);
  signal fifo1_read_toggle : std_logic := '0';

  signal fifo1_f : std_logic;
  signal fifo1_nf : std_logic;
  signal fifo1_write : std_logic;
  signal fifo1_q : std_logic_vector(63 downto 0);
  signal fifo1_data : std_logic_vector(63 downto 0);
  signal afifo1_empty : std_logic;

  
  signal rst_i : std_logic;
  signal pcie_clk : std_Logic;  
  
  component small_sync_fifo
    port (
      clk: IN std_logic;
      din: IN std_logic_VECTOR(63 downto 0);
      rd_en: IN std_logic;
      wr_en: IN std_logic;
      dout: OUT std_logic_VECTOR(63 downto 0);
      empty: OUT std_logic;
      full: OUT std_logic);
  end component;


  signal tohostfifo_enable : std_logic;
  signal tofpgafifo_enable : std_logic;
  signal tohost_count : std_logic_vector(15 downto 0);
  signal tofpga_count : std_logic_vector(15 downto 0);
  


begin  -- rtl

  
  -- 64 - 128 bit conversion required
  -- Put conversion at input of write FIFO
  -- and between read FIFOs
  
  process (pcie_clk, rst_i)
  begin  -- process
    if rst_i = '1' then   
      fifo0_write_toggle <= '0';
      fifo0_ld_iq_reg <= (others => '0');       
      fifo1_read_toggle <= '0';
    elsif pcie_clk'event and pcie_clk = '1' then  -- rising clock edge
      if fifo0_write = '1' then
        fifo0_write_toggle <= not fifo0_write_toggle;
        fifo0_ld_iq_reg <= tofpga_fifo_data1(63 downto 0);
      --fifo0_ld_iq_reg(47 downto 32) <= X"1234";
      end if;  
      if fifo1_write = '1' then
        fifo1_read_toggle <= not fifo1_read_toggle;
      end if;
      fifo0_nfull_reg <= fifo0_nf;
    end if;
  end process;        

  tofpga_fifo_read1 <= not fifo0_nfull_reg and not tofpga_fifo_empty1;
  fifo0_write <= tofpga_fifo_read;
  
  fifo0_ld_iq <= tofpga_fifo_data1(63 downto 0) & fifo0_ld_iq_reg;   
  fifo0_write_q <= fifo0_write and fifo0_write_toggle;
  
  -- Always write 2 words into output FIFO
  fifo1_write <= (not afifo1_empty and not fifo1_nf) or fifo1_read_toggle;
  fifo1_data(63 downto 0) <= afifo1_qdata(63 downto 0) when fifo1_read_toggle = '0' else afifo1_qdata(127 downto 64);
  --fifo1_data(47 downto 0) <= afifo1_qdata(47 downto 0) when fifo1_read_toggle = '0' else afifo1_qdata(111 downto 64);
  --fifo1_data(63 downto 48) <= X"ABCD";
  afifo1_read <= fifo1_write and fifo1_read_toggle;
  
  afifo0 : afifo
    generic map(
      width => 128,
      order => 9,
      wpfl => 3,
      wpfl2 => 0,
      rpel => 2,
      family => family,
      safety => false,
      fflush => false)
    port map(
      rst => rst_i,
      wk => pcie_clk,
      wd => fifo0_ld_iq,
      wadv => fifo0_write_q,
      we => open,
      wpf => fifo0_nf,
      wpf2 => open,
      wf => open,
      wovf => open,
      wlevel => open,
      rk => fifo_clk,
      rq => rfifo0_outputs.data,
      radv => rfifo0_inputs.rd,
      rflush => '0',
      re => rfifo0_outputs.empty,
      rpe => rfifo0_outputs.nempty,
      rf => open,
      rudf => open,
      rlevel => rfifo0_outputs.level);

  
  sfifo1 : sfifo
    generic map(
      width      => 64,
      order      => 5,
      rpel       => 2,
      wpfl       => 3,
      family     => family,
      safety     => false)
    port map(
      rst        => rst_i,
      k          => pcie_clk,
      wd         => fifo1_data,
      wadv       => fifo1_write,
      we         => open,
      wpf        => fifo1_nf,
      wf         => fifo1_f,
      wovf       => open,
      rq         => tohost_fifo_data1,
      radv       => tohost_fifo_read1,
      rflush     => '0',
      re         => tohost_fifo_empty1,
      rf         => open,
      rpe        => open,
      rudf       => open,
      rlevel     => open);

  
  
  afifo1 : afifo
    generic map(
      width => 128,
      order => 9,
      wpfl => 3,
      wpfl2 => 0,
      rpel => 2,
      family => family,
      safety => false,
      fflush => false)
    port map(
      rst => rst_i,
      wk => fifo_clk,
      wd => wfifo1_inputs.data,
      wadv => wfifo1_inputs.wr,
      we => open,
      wpf => wfifo1_outputs.nfull,
      wpf2 => open,
      wf => wfifo1_outputs.full,
      wovf => open,
      wlevel => wfifo1_outputs.level,
      rk => pcie_clk,
      rq => afifo1_qdata,
      radv => afifo1_read,
      rflush => '0',
      re => afifo1_empty,
      rpe => open,
      rf => open,
      rudf => open,
      rlevel => open);

  -- Small Synchronous FIFOs to help timing
--     ssf0: small_sync_fifo
-- 		port map (
-- 			clk => pcie_clk,
-- 			din => tofpga_fifo_data,
-- 			rd_en => tofpga_fifo_read1,
-- 			wr_en => tofpga_fifo_read,
-- 			dout => tofpga_fifo_data1,
-- 			empty => tofpga_fifo_empty1,
-- 			full => ssf0_full);
--          
-- 		tofpga_fifo_read <= not ssf0_full and not tofpga_fifo_empty;	
-- 			   
--     ssf1: small_sync_fifo
-- 		port map (
-- 			clk => pcie_clk,
-- 			din => tohost_fifo_data1,
-- 			rd_en => tohost_fifo_read,
-- 			wr_en => tohost_fifo_read1,
-- 			dout => tohost_fifo_data,
-- 			empty => tohost_fifo_empty,
-- 			full => ssf1_full);
-- 			     
-- 			tohost_fifo_read1 <= not ssf1_full and not tohost_fifo_empty1;	
  
  
  tohost_fifo_read1  <= tohost_fifo_read;
  tohost_fifo_data <= tohost_fifo_data1 when tohostfifo_enable = '1' else X"DEADF1F0" & tofpga_count & tohost_count;
  tohost_fifo_empty  <= tohost_fifo_empty1 when tohostfifo_enable = '1' else '0';
  tofpga_fifo_data1  <= tofpga_fifo_data;
  tofpga_fifo_empty1 <= tofpga_fifo_empty;
  tofpga_fifo_read   <= tofpga_fifo_read1 when tofpgafifo_enable = '1' else not tofpga_fifo_empty;
  
--     process (pcie_clk, rst_i)
--     begin  -- process
--       if rst_i = '1' then   
--         tohost_fifo_data(15 downto 0) <= (others => '0');       
--       elsif pcie_clk'event and pcie_clk = '1' then  -- rising clock edge
--         if tohost_fifo_read = '1' then
--           tohost_fifo_data(15 downto 0) <= tohost_fifo_data(15 downto 0)+1;
--         end if;  
--       end if;
--     end process;    
  
  
  pcie_core0: slimbridge
    port map (
      sys_clk_p          => pcie_clk_p,
      sys_clk_n          => pcie_clk_n,
      sys_reset_n        => pcie_rst_n,
      pci_exp_txp         => pcie_tx_p,
      pci_exp_txn         => pcie_tx_n,
      pci_exp_rxp         => pcie_rx_p,
      pci_exp_rxn         => pcie_rx_n,
      id                => identifier,
      reg_addr => slave_bus_in.a(9 downto 2),
      reg_data => slave_bus_out.q(31 downto 0),
      
      fifo_clk => pcie_clk,
      fifo_rst => rst_i,
      tofpga_fifo_data  => tofpga_fifo_data,
      tofpga_fifo_empty => tofpga_fifo_empty,
      tofpga_fifo_read  => tofpga_fifo_read,
      tohost_fifo_data  => tohost_fifo_data,
      tohost_fifo_empty => tohost_fifo_empty,
      tohost_fifo_read  => tohost_fifo_read);

  slave_bus_out.qv <= slave_bus_in.rd;

  -- Automatically read and write PCIe FIFOs if connection not enabled.
  fifo_debug_mode: process (pcie_clk, rst_i)
  begin  -- process fifo_debug_mode
    if rst_i = '1' then                 -- asynchronous reset (active low)
      tohostfifo_enable <= '0';
      tofpgafifo_enable <= '0';
      tohost_count <= (others => '0');
      tofpga_count <= (others => '0');
    elsif pcie_clk'event and pcie_clk = '1' then  -- rising clock edge
      if tohostfifo_enable = '1' then
        tohost_count <= (others => '0');
        tofpga_count <= (others => '0');
      else
        if tofpga_fifo_read = '1' then
          tofpga_count <= tofpga_count+1;
        end if;
        if tohost_fifo_read = '1' then
          tohost_count <= tohost_count+1;
        end if;
      end if;
      tohostfifo_enable <= identifier(0);
      tofpgafifo_enable <= identifier(1);
    end if;
  end process fifo_debug_mode;
  
  
end rtl;
